/*
** main.c
**
** Copyright (c) 1994-1995 Peter Eriksson <pen@signum.se>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <thread.h>
#include <errno.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>
#include <netdb.h>
#include <fcntl.h>
#include <syslog.h>

#include <sys/systeminfo.h>

#include <sys/wait.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "phttpd.h"

#ifndef INSTROOT
#  define INSTROOT "/opt/www"
#endif

#define CONFIG_FILE     INSTROOT "/etc/phttpd.conf"

#define DEFAULT_LOGFILE "/var/log/phttpd/requests"
#define DEFAULT_PIDFILE "/var/log/phttpd/pid"
#define DEFAULT_ERRFILE "/var/log/phttpd/stderr"
#define DEFAULT_CNTFILE "/var/log/phttpd/counts"

#define DEFAULT_ADDR    "*"
#define DEFAULT_PORT	80
#define DEFAULT_HOME    INSTROOT "/db"
#define DEFAULT_UID     60001
#define DEFAULT_GID     60001

#define DEFAULT_PROCESSES      0
#define DEFAULT_ACCEPT_THREADS 0

#define DEFAULT_MODHOME INSTROOT "/modules"

#define DEFAULT_FILE_HANDLER "file.so"
#define DEFAULT_DIR_HANDLER  "dir.so"



static struct connectioninfo **ciptab;

static mutex_t req_lock;

static mutex_t *accept_lock_p = NULL;

int extended_logging = 0;
int debug = 0;
int n_requests = 0;
int n_accept = DEFAULT_ACCEPT_THREADS;
int n_processes = DEFAULT_PROCESSES;
int wait_mode = 0;
int no_copying_info = 0;

int listen_sock = -1;
#if defined(__i386__) && defined(__sun__)
int stack_size = 128*1024;
#else
int stack_size = 0;
#endif

int keepalive_enabled = 1;
int keepalive_timeout = 30;
int keepalive_maxreq = 4;


char *user_dir    = ".public";

char *server_host = NULL;
char *server_addr = DEFAULT_ADDR;
int   server_port = DEFAULT_PORT;
char *server_home = DEFAULT_HOME;
char *modules_home = DEFAULT_MODHOME;
int   bound_threads = 0;
int   hostname_lookups = 0;
int   ident_lookups = 0;


char *web_admin_name = NULL;
char *web_admin_home = NULL;
char *web_admin_email = NULL;

char *default_file_handler = "file.so";
char *default_dir_handler  = "dir.so";

char server_url[2048] = "<not set>";

char *server_user = NULL;
char *server_group = NULL;

int   server_uid  = DEFAULT_UID;
int   server_gid  = DEFAULT_GID;

int concurrency = 20;

/* NOTE! See the information in the INSTALL file about this n_listen
   value! You must modify your kernel (according to the INSTALL file)
   to be able to set this value to anything above 5 and have the
   operating system acknowledge it! */
int n_listen = 1024; 
		      

int gc_time  = 2*60;   /* Terminate hanging connections after second minute */
int gc_sleep = 1*60;   /* Check for hanging connections every minute */

sema_t reconfig_sema;

time_t start_time;

char *logfile_path = NULL;
char *pidfile_path = NULL;
char *cntfile_path = NULL;
int log_fd = -1;

char *errfile_path = NULL;
char *less_path = NULL;


static struct options global_cfg_table[] =
{
    { "url-handlers",      T_TABLE, &url_handlers_table,	NULL, 0, 0, NULL },
    { "url-redirects",     T_TABLE, &url_redirects_table,	NULL, 0, 0, NULL },
    { "url-permanent-redirects",     T_TABLE, &url_predirects_table,	NULL, 0, 0, NULL },
    { "url-rewrites",      T_TABLE, &url_rewrites_table,	NULL, 0, 0, NULL },
    { "url-authenticate",  T_TABLE, &url_authenticate_table,    NULL, 0, 0, NULL },
    { "url-access",        T_TABLE, &url_access_table,          NULL, 0, 0, NULL },
    { "url-expand",	   T_TABLE, &url_expand_table,		NULL, 0, 0, NULL },

    { "config-file",	   T_FILE,   NULL,			NULL, 'f', 0, NULL, "Path to config file" },
    
    { "database-dir",      T_STRING, &server_home,		NULL, 0, 0, NULL, "Path to database directory" },
    { "server-name",       T_STRING, &server_host,		NULL, 0, 0, NULL, "Official FQDN name of server" },
    { "server-addr",       T_STRING, &server_addr,		NULL, 0, 0, NULL, "Address to bind server to" },
    { "modules-dir",	   T_STRING, &modules_home,             NULL, 0, 0, NULL, "Path to modules directory" },
    { "user-dir",          T_STRING, &user_dir,			NULL, 0, 0, NULL, "Users public directory" },
    { "logging-file",      T_STRING, &logfile_path,		NULL, 'L', 0, DEFAULT_LOGFILE, "Path to requests logging file" },
    { "counting-file",     T_STRING, &cntfile_path,             NULL, 'C', 0, DEFAULT_CNTFILE, "Path to requests counter database" },
    { "pid-file",          T_STRING, &pidfile_path,             NULL, 'P', 0, DEFAULT_PIDFILE, "Path to PID file" },
    { "stderr-file",       T_STRING, &errfile_path,             NULL, 'E', 0, DEFAULT_ERRFILE, "Path to stderr log file" },
    { "LESS-socket",	   T_STRING, &less_path,		NULL, 0, 0, NULL,   "Path to LESS socket" },
    
    { "web-admin-name",    T_STRING, &web_admin_name,           NULL, 0, 0, NULL,  "Webmaster's name" },
    { "web-admin-home",    T_STRING, &web_admin_home,           NULL, 0, 0, NULL,  "Webmaster's home URL" },
    { "web-admin-email",   T_STRING, &web_admin_email,          NULL, 0, 0, NULL,  "Webmaster's email address" },
    
    { "server-user",       T_STRING, &server_user,		NULL, 'u', 0, NULL, "Server user name or number" },
    { "server-group",      T_STRING, &server_group,		NULL, 'g', 0, NULL, "Server group name or number" },

    { "debug-level",	   T_NUMBER, &debug,			NULL, 'd', 1, NULL, "Debug level" },

    { "fscache-size",      T_NUMBER, &fscache_size,             NULL, 0, 0, NULL, "Fscache size" },
    { "fscache-refresh",   T_NUMBER, &fscache_refresh,          NULL, 0, 0, NULL, "Fscache refresh interval" },
    { "fscache-ttl",       T_NUMBER, &fscache_ttl,              NULL, 0, 0, NULL, "Fscache time-to-live" },
    { "fscache-gc-interval", T_NUMBER, &fscache_gc_interval,    NULL, 0, 0, NULL, "Fscache garbage collect interval" },

    { "hostcache-size",      T_NUMBER, &hostcache_size,             NULL, 0, 0, NULL, "Hostcache size" },
    { "hostcache-refresh",   T_NUMBER, &hostcache_refresh,          NULL, 0, 0, NULL, "Hostcache refresh interval" },
    { "hostcache-ttl",       T_NUMBER, &hostcache_ttl,              NULL, 0, 0, NULL, "Hostcache time-to-live" },
    { "hostcache-gc-interval", T_NUMBER, &hostcache_gc_interval,    NULL, 0, 0, NULL, "Hostcache garbage collect interval" },

    { "usercache-size",      T_NUMBER, &usercache_size,             NULL, 0, 0, NULL, "Usercache size" },
    { "usercache-refresh",   T_NUMBER, &usercache_refresh,          NULL, 0, 0, NULL, "Usercache refresh interval" },
    { "usercache-ttl",       T_NUMBER, &usercache_ttl,              NULL, 0, 0, NULL, "Usercache time-to-live" },
    { "usercache-gc-interval", T_NUMBER, &usercache_gc_interval,    NULL, 0, 0, NULL, "Usercache garbage collect interval" },

    { "urlcache-size",      T_NUMBER, &urlcache_size,             NULL, 0, 0, NULL, "Urlcache size" },
    { "urlcache-refresh",  T_NUMBER, &urlcache_refresh,         NULL, 0, 0, NULL, "Urlcache refresh interval" },
    { "urlcache-ttl",      T_NUMBER, &urlcache_ttl,             NULL, 0, 0, NULL, "Urlcache time-to-live" },
    { "urlcache-gc-interval", T_NUMBER, &urlcache_gc_interval,     NULL, 0, 0, NULL, "Urlcache garbage collect interval" },
    
    { "server-port",       T_NUMBER, &server_port,		NULL, 'p', 8000, NULL, "Servers' TCP port" },
    { "server-processes",  T_NUMBER, &n_processes,              NULL, 'Q', 1, NULL, "Concurrent accepting processes" },
    { "accept-threads",    T_NUMBER, &n_accept,                 NULL, 'A', 1, NULL, "Concurrent accepting threads" },
    { "concurrency",       T_NUMBER, &concurrency,		NULL, 0, 0, NULL,   "Threads concurrency level" },
    { "listen-backlog",    T_NUMBER, &n_listen,			NULL, 'b', 1024, NULL, "Listen backlog level" },
    { "socket-sndbuf",     T_NUMBER, &so_sndbuf, 	        NULL, 0, 0, NULL, "Size of socket send buffer" },
    { "socket-rcvbuf",     T_NUMBER, &so_rcvbuf, 	        NULL, 0, 0, NULL, "Size of socket receive buffer" },
    { "gc-time",           T_NUMBER, &gc_time,			NULL, 0, 0, NULL, "Connection time-to-live" },
    { "gc-sleep",          T_NUMBER, &gc_sleep,			NULL, 0, 0, NULL, "Connection garbage collect interval" },
    { "select-timeout",    T_NUMBER, &select_timeout,		NULL, 'T', 0, NULL, "Select timeout limit" },
    { "stack-size",        T_NUMBER, &stack_size,               NULL, 0, 0, NULL, "Threads stack size" },

    { "extended-logging",  T_BOOL, &extended_logging,		NULL, 0, 1, NULL, "Enable extended logging" },
    { "keepalive-connections", T_BOOL, &keepalive_enabled,      NULL, 0, 1, NULL, "Enable Keep-Alive connections" },
    { "bound-threads",	   T_BOOL,   &bound_threads,		NULL, 'B', 1, NULL, "Use bound threads" },
    { "hostname-lookups",  T_BOOL,   &hostname_lookups,         NULL, 'H', 1, NULL, "Perform hostname lookups" },
    { "ident-lookups",  T_BOOL,   &ident_lookups,         NULL, 'I', 1, NULL, "Perform IDENT lookups" },
    { "no-copying-info",   T_BOOL,   &no_copying_info,          NULL, 's', 1, NULL, "Don't display the copying info" },
    { "wait-mode",	   T_BOOL,   &wait_mode,		NULL, 'w', 1, NULL, "Start server in 'wait' mode" },
    
    { NULL,                -1,       NULL,			NULL, 0, 0, NULL, NULL }
};



static int reload_module(struct modinfo *mp, void *misc)
{
    mutex_lock(&mp->lock);
    mp->update = 1;
    mutex_unlock(&mp->lock);
    
    return 0;
}

static void stderr_init(char *path)
{
    int fd;


    if (debug)
	return;
    
    if (path == NULL)
        path = "/dev/null";

    fd = s_open(path, O_WRONLY+O_APPEND+O_CREAT, 0644);
    if (fd >= 0 && fd != 2)
    {
	s_dup2(fd, 2);
	s_close(fd);
    }
}

static void *reconfig_thread(void *misc)
{
  Loop:
    s_sema_wait(&reconfig_sema);
    syslog(LOG_NOTICE, "SIGHUP received - reconfiguring");

    log_reopen();
    stderr_init(errfile_path);
    
    md_foreach(reload_module, NULL);
    goto Loop;
}


/* XXX: Needs to be rewritten & rethought! Should terminate gracefully */
static void *timeout_thread(int *timeout)
{
    s_sleep(*timeout);
    
    _exit(0);
    return NULL;
}


static int gc_conn(struct connectioninfo *cip, void *misc)
{
    time_t curr_time;


    curr_time = * (time_t *) misc;

    
    if (curr_time - cip->cn_time > gc_time)
    {
	syslog(LOG_DEBUG, "terminating connection #%d", cip->request_no);
	fd_shutdown(cip->fd, -2);
    }
 
    return 0;
}


/* The Garbage Collector Thread - used to clean up hanging connections */
static void *gc_thread(void *misc)
{
    time_t t;


  Loop:
    s_sleep(gc_sleep);

    if (debug > 1)
	write(2, "gc_thread(): Timeout\n", 21);

    time(&t);
    clist_foreach(gc_conn, (void *) &t);
    goto Loop;
}

static void sighup_handler()
{
    sema_post(&reconfig_sema);
}


static void sigusr1_handler()
{
    debug++;
}

static void sigusr2_handler()
{
    debug--;
}


static void sigterm_handler()
{
    syslog(LOG_NOTICE, "server terminating by SIGTERM request");
    exit(0);
}



static char *addr2host(struct sockaddr_in *sinp)
{
    struct hostent *hep;
    void *key;
    char *result;
    

    if ((key = hostcache_lookup_byaddr(&sinp->sin_addr, 0, &hep)) != NULL)
    {
	if (hep && hep->h_name)
	    result = s_strdup(hep->h_name);
	else
	    result = NULL;
	    
	hostcache_release(key);

	return result;
    }
    
    return NULL;
}

static void *hostname_thread(struct addressinfo_s *aip)
{
    char *host;

    
    host = addr2host(&aip->addr);
    
    mutex_lock(&aip->host.mtx);
    aip->host.name = (host ? host : s_strdup(""));
    cond_broadcast(&aip->host.cv);
    mutex_unlock(&aip->host.mtx);

    return NULL;
}


static void *ident_thread(struct addressinfo_s *aip)
{
    int fd;
    struct sockaddr_in sin;
    char buf[2048];
    char *name = NULL;
    char *cp;

    
    sin = aip->addr;
    sin.sin_port = htons(113);
    
    fd = fd_connect((struct sockaddr *) &sin, sizeof(sin));
    if (fd < 0)
	goto Done;
    
    fd_printf(fd, "%d , %d\n", aip->port, server_port);
    fd_shutdown(fd, 1);

    if (fd_gets(buf, sizeof(buf), fd) == NULL)
	goto Done;
    
    name = strtok_r(buf, ":", &cp);
    if (name)
    {
	/* Got "port,port" pair */
	
	name = strtok_r(NULL, " \t\r\n:", &cp);
	if (name)
	{
	    /* Got "USERID" or "ERROR" */
	    
	    if (strcasecmp(name, "USERID") != 0)
	    {
		name = NULL;
		goto Done;
	    }
	    
	    name = strtok_r(NULL, " \t\r\n:", &cp);
	    if (name)
	    {
		/* Got OS-name */

		name = strtok_r(NULL, "\r\n", &cp);
 		if (name)
		{
		    /* Got ident identifier */

		    goto Done;
		}
	    }
	}
    }
    
    name = NULL;

  Done:
    if (fd >= 0)
	fd_close(fd);
    
    mutex_lock(&aip->ident.mtx);
    aip->ident.name = s_strdup(name ? name : "");
    cond_broadcast(&aip->ident.cv);
    mutex_unlock(&aip->ident.mtx);

    return NULL;
}



void start_hostinfo_lookup(struct addressinfo_s *aip,
			   int do_ident)
{
    mutex_init(&aip->host.mtx, USYNC_THREAD, NULL);
    cond_init(&aip->host.cv, USYNC_THREAD, NULL);

    if (hostname_lookups)
    {
	aip->host.name = NULL;

	if (thr_create(NULL,
		       stack_size,
		       (void *(*)(void *)) hostname_thread,
		       (void *) aip,
		       THR_DETACHED+THR_DAEMON,
		       NULL))
	{
	    syslog(LOG_ERR, "thr_create(hostname_thread) failed: %m");
	    exit(1);
	}
    }
    else
	aip->host.name = s_strdup("");

    mutex_init(&aip->ident.mtx, USYNC_THREAD, NULL);
    cond_init(&aip->ident.cv, USYNC_THREAD, NULL);
    
    if (ident_lookups && do_ident)
    {
	aip->ident.name = NULL;

	if (thr_create(NULL,
		       stack_size,
		       (void *(*)(void *)) ident_thread,
		       (void *) aip,
		       THR_DETACHED+THR_DAEMON,
		       NULL))
	{
	    syslog(LOG_ERR, "thr_create(ident_thread) failed: %m");
	    exit(1);
	}
    }
    else
	aip->ident.name = s_strdup("");
}


char *get_hostname(struct addressinfo_s *aip)
{
    if (!hostname_lookups)
	return NULL;
    
    mutex_lock(&aip->host.mtx);
    while (aip->host.name == NULL) {
	cond_wait(&aip->host.cv, &aip->host.mtx);
    }
    mutex_unlock(&aip->host.mtx);


    if (debug > 1)
	fprintf(stderr, "get_hostname hostname=%s\n", aip->host.name);
    
    return aip->host.name[0] ? aip->host.name : NULL;
}


char *get_ident(struct addressinfo_s *aip)
{
    if (!ident_lookups)
	return NULL;
    
    mutex_lock(&aip->ident.mtx);
    while (aip->ident.name == NULL) {
	cond_wait(&aip->ident.cv, &aip->ident.mtx);
    }
    mutex_unlock(&aip->ident.mtx);

    if (debug > 1)
	fprintf(stderr, "get_ident username=%s\n", aip->ident.name);
    
    return aip->ident.name[0] ? aip->ident.name : NULL;
}





static void *request_thread(struct connectioninfo *cip)
{
    int len, result;
    int keepalive = 0;
    

    if (debug > 1)
	fprintf(stderr, "request_thread(): Start\n");
    
    cip->tid = thr_self();

    cip->client.addr_s = s_strdup(inet_ntoa(cip->client.addr.sin_addr));
    cip->client.port = ntohs(cip->client.addr.sin_port);

    len = sizeof(cip->server.addr);
    
    if (getsockname(cip->fd,
		    (struct sockaddr *) &cip->server.addr,
		    &len) < 0)
    {
	syslog(LOG_INFO, "getsockname() failed: %m");
	cip->server.addr_s = NULL;
    }
    else
	cip->server.addr_s = s_strdup(inet_ntoa(cip->server.addr.sin_addr));

    start_hostinfo_lookup(&cip->server, 0);
    start_hostinfo_lookup(&cip->client, 1);
    
    time(&cip->cn_time);

    clist_insert(cip);
    
    fd_reopen(cip->fd, 2, FDT_SOCKET);

  Again:
    keepalive = 0;
    result = http_get_request(cip, &keepalive);

    if (keepalive_enabled && keepalive && result > 0)
    {
	if (debug > 1)
	    fprintf(stderr, "Keepalive, retrying (fd=%d)\n", cip->fd);

	fd_flush(cip->fd);
	
	if (cip->hip)
	{
	    log_connection(cip, result);
	    http_freeinfo(cip->hip);
	    cip->hip = NULL;
	}

	goto Again;
    }
    
    if (debug > 1)
	fprintf(stderr, "No keepalive, closing (fd=%d)\n", cip->fd);

    clist_delete(cip);
    fd_shutdown(cip->fd, 2);

    if (cip->hip)
    {
	log_connection(cip, result);
	http_freeinfo(cip->hip);
	cip->hip = NULL;
    }
    
    if (cip->server.addr_s != NULL)
	s_free(cip->server.addr_s);

    if (cip->client.addr_s != NULL)
	s_free(cip->client.addr_s);

    (void) get_hostname(&cip->server);
    (void) get_hostname(&cip->client);
    
    if (cip->server.host.name != NULL)
	s_free(cip->server.host.name);

    if (cip->client.host.name != NULL)
	s_free(cip->client.host.name);

    (void) get_ident(&cip->server);
    (void) get_ident(&cip->client);
    
    if (cip->server.ident.name != NULL)
	s_free(cip->server.ident.name);

    if (cip->client.ident.name != NULL)
	s_free(cip->client.ident.name);

    if (cip->auth.type != NULL)
	s_free(cip->auth.type);

    if (cip->auth.user != NULL)
	s_free(cip->auth.user);

    fd_close(cip->fd);
    
    if (debug > 1)
	fprintf(stderr, "request_thread(): Stop\n");
    
    return NULL;
}


/* LESS - Lysator Efficient Server-Side Redirect */
void *less_thread(void *sock)
{
    struct iovec  iov;
    struct msghdr msg;
    int sock_fd, len;
    struct connectioninfo *cip;

    sock_fd = * (int *) sock;

    
  Loop:
    iov.iov_base         = NULL;
    iov.iov_len          = 0;
    msg.msg_iov          = &iov;
    msg.msg_iovlen       = 1;
    msg.msg_name         = NULL;
    msg.msg_accrights    = (char *) &sock_fd;
    msg.msg_accrightslen = sizeof(sock_fd);
    
    if (recvmsg(sock_fd, &msg, 0) < 0)
    {
	syslog(LOG_INFO, "recvmsg() failed: %m");
	goto Loop;
    }

    if (msg.msg_iov->iov_len > 0)
    {
	/* Got some preparsed data (METHOD URL VERSION\nMIME-Headers)
	   which we need to handle before calling the normal handler */

	/* (Perhaps one can insert it into the fdstdio input buffer?) */
	
	/* XXXX - TO BE DONE - XXXX */
    }
    
    if (ciptab[sock_fd] == NULL)
	cip = ciptab[sock_fd] = s_malloc(sizeof(struct connectioninfo));
    else
	memset(cip = ciptab[sock_fd], 0, sizeof(struct connectioninfo));
    
    cip->fd = sock_fd;
    len = sizeof(cip->client.addr);
    
    if (getsockname(sock_fd,
		    (struct sockaddr *) &cip->client.addr,
		    &len) < 0)
    {
	syslog(LOG_INFO, "getsockname(): %m");
	s_close(sock_fd);
	goto Loop;
    }
    

    if (thr_create(NULL,
		   stack_size,
		   (void *(*)(void *)) request_thread,
		   (void *) cip,
		   THR_DETACHED+THR_DAEMON+(bound_threads ? THR_BOUND : 0),
		   NULL))
    {
	syslog(LOG_ERR, "thr_create(request_thread) failed: %m");
	exit(1);
    }
    
    goto Loop;
}



static void *accept_thread(void *sock)
{
    int len, listen_sock;
    struct connectioninfo *cip;
    int fd;
    struct sockaddr_in sin;
    

    listen_sock = * (int *) sock;

  Loop:
    len = sizeof(sin);
    memset(&sin, 0, len);

    if (accept_lock_p)
	mutex_lock(accept_lock_p);
    
    fd = s_accept(listen_sock, (struct sockaddr *) &sin, &len);

    if (accept_lock_p)
	mutex_unlock(accept_lock_p);

    if (fd < 0)
    {
	syslog(LOG_INFO, "accept() failed: %m");
	goto Loop;
    }

    if (ciptab[fd] == NULL)
	cip = ciptab[fd] = s_malloc(sizeof(struct connectioninfo));
    else
	memset(cip = ciptab[fd], 0, sizeof(struct connectioninfo));
    
    cip->fd = fd;
    cip->client.addr = sin;

    mutex_lock(&req_lock);
    cip->request_no = ++n_requests;
    mutex_unlock(&req_lock);

    if (thr_create(NULL,
		   stack_size,
		   (void *(*)(void *)) request_thread,
		   (void *) cip,
		   THR_DETACHED+THR_DAEMON+(bound_threads ? THR_BOUND : 0),
		   NULL))
    {
	syslog(LOG_ERR, "thr_create(request_thread) failed: %m");
	exit(1);
    }
    

    goto Loop;
}    


static char *get_local_hostname(char *buf,
				int size)
{
    struct hostent *hp;
    int i;

    
    if (sysinfo(SI_HOSTNAME, buf, size) < 0)
	return NULL;

    if (strchr(buf, '.') != NULL)
	return buf;

    hp = gethostbyname(buf);
    if (hp == NULL)
    {
	syslog(LOG_WARNING, "unable to find FQDN for %s", buf);
	return buf;
    }
    
    if (hp->h_name && strchr(hp->h_name, '.') != NULL)
    {
	if ((int) strlen(hp->h_name) < size)
	{
	    strcpy(buf, hp->h_name);
	    return buf;
	}

	/* else name too long, try to find a shorter in the aliases list */
    }

    if (hp->h_aliases)
	for (i = 0; hp->h_aliases[i]; i++)
	{
	    if (strchr(hp->h_aliases[i], '.') != NULL)
	    {
		if (size > (int) strlen(hp->h_aliases[i]))
		{
		    strcpy(buf, hp->h_aliases[i]);
		    return buf;
		}
		
		/* else name too long, try to find a shorter one later */
	    }
	}
    
    syslog(LOG_WARNING, "unable to find FQDN for %s", buf);
    return buf;
}


static int get_addr(char *host, struct in_addr *ia)
{
    if (isdigit(host[0]))
	ia->s_addr = inet_addr(host);
    else
    {
	struct hostent hp;
	int h_errno;
	char buf[2048];
	

	if (s_gethostbyname_r(host, &hp, buf, sizeof(buf), &h_errno) == NULL)
	    return -1;

	memcpy(ia, hp.h_addr_list[0], hp.h_length);
    }

    return 0;

}

int uidgid_get(const char *user,
	       const char *group,
	       int *uid,
	       int *gid,
	       struct passwd **pwp,
	       struct passwd *pwb,
	       char *pbuf,
	       int pbuflen)
{
    struct group *grp, grb;
    char gbuf[1024];
    
    
    if (user != NULL)
    {
	if (isdigit(user[0]))
	{
	    *uid = atoi(user);
	    *pwp = s_getpwuid_r(*uid, pwb, pbuf, pbuflen);
	}
	else
	{
	    *pwp = s_getpwnam_r(user, pwb, pbuf, pbuflen);

	    if (*pwp == NULL)
		return -1;
	    
	    *uid = (*pwp)->pw_uid;
	}

	if (*pwp)
	    *gid = (*pwp)->pw_gid;
    }


    if (group != NULL)
    {
	if (isdigit(group[0]))
	    *gid = atoi(group);
	else
	{
	    grp = s_getgrnam_r(group, &grb, gbuf, sizeof(gbuf));
	    if (grp == NULL)
		return -2;
	    
	    *gid = grp->gr_gid;
	}
    }

    return 0;
}


void start_less(void)
{
    int less_sock;
    struct sockaddr_un sun_sa;
    
    
    if (less_path == NULL)
	return;
    
    if ((less_sock = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0)
	return;

    memset(&sun_sa, 0, sizeof(sun_sa));
    sun_sa.sun_family = AF_UNIX;
    strncpy(sun_sa.sun_path, less_path, sizeof(sun_sa.sun_path));
    
    if (bind(less_sock, (struct sockaddr *) &sun_sa, sizeof(sun_sa)) < 0)
    {
	s_close(less_sock);
	return;
    }
    
    if (thr_create(NULL,
		   stack_size,
		   (void *(*)(void *)) less_thread,
		   (void *) &less_sock,
		   THR_DETACHED+THR_DAEMON+THR_BOUND,
		   NULL))
    {
	syslog(LOG_ERR, "thr_create(less_thread) failed: %m");
	s_close(less_sock);
	exit(1);
    }
}


int main(int argc,
	 const char *argv[])
{
    int i, status, timeout;
    int one = 1;
    struct sockaddr_in sin;
    struct sigaction actb;
    struct passwd *pwp, pwb;
    char pbuf[2048];
    static char host_name[1024];
    

    openlog("phttpd", LOG_PID, LOG_DAEMON);
    
    time(&start_time);
    
    timeout = 0;
    pwp = NULL;
    
    sema_init(&reconfig_sema, 0, USYNC_THREAD, NULL);

    for (i = 1; i < argc && argv[i][0] == '-'; i++)
	if (argv[i][1] == 'd')
	    if (argv[i][2])
		debug = atoi(argv[i]+2);
	    else
		debug = 1;

    fd_init();
    
    config_parse_file(CONFIG_FILE, global_cfg_table, 0);

    i = config_parse_argv(argv+1, global_cfg_table);
    if (i++ < 0)
	exit(1);
    
    if (wait_mode)
    {
	listen_sock = 0;

	if (debug == 0)
	    debug = -1;
    }	

    if (n_accept == 0)
	n_accept = sysconf(_SC_NPROCESSORS_ONLN);
    
    if (uidgid_get(server_user, server_group,
		   &server_uid, &server_gid,
		   &pwp, &pwb, pbuf, sizeof(pbuf)) < 0)
    {
	syslog(LOG_ERR, "uidgid_get(\"%s\", \"%s\") failed",
	       server_user != NULL ? server_user : "",
	       server_group != NULL ? server_group : "");
	exit(1);
    }
    
    if (server_host == NULL)
    {
	if (get_local_hostname(host_name, sizeof(host_name)) == NULL)
	{
	    syslog(LOG_ERR, "get_local_hostname() failed: %m");

	    fprintf(stderr, "%s: get_local_hostname(): ", argv[0]);
	    perror("");
	    exit(1);
	}

	server_host = host_name;
    }

    if (server_port == 80)
	sprintf(server_url, "http://%s", server_host);
    else
	sprintf(server_url, "http://%s:%d", server_host, server_port);
	
    if (listen_sock < 0)
    {
	printf("[Phttpd %s - Copyright (c) 1995 Peter Eriksson <pen@signum.se>]\n\n",
	       server_version);

	if (!no_copying_info)
	{
	    puts("\tPhttpd comes with ABSOLUTELY NO WARRANTY. This is free software,");
	    puts("\tand you are welcome to redistribute it under certain conditions;");
	    puts("\tfor more information about the conditions and non-warranty, see:\n");
	    printf("\t\t%s/phttpd/COPYING.html\n\n",
		       server_url);
	    
	    puts("\tor read the file COPYING that should have been part of the");
	    puts("\tsoftware distribution - if not send an email to <pen@signum.se>.\n");
	}
    }
    

    if (i < argc)
	server_home = s_strdup(argv[i]);
    
    if (server_home)
    {
	status = s_chdir(server_home);
	if (status < 0)
	{
	    syslog(LOG_ERR, "chdir(\"%s\") failed: %m", server_home);
	    
	    fprintf(stderr, "%s: chdir(\"%s\"): ", argv[0], server_home);
	    perror("");
	    exit(1);
	}
    }

    if (listen_sock == -1)
    {
	if ((listen_sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
	    syslog(LOG_ERR, "socket() failed: %m");
	    
	    fprintf(stderr, "%s: socket(AF_INET, SOCK_STREAM, 0): ",
		    argv[0]);
	    perror("");
	    exit(1);
	}
	
	/* We ignore any error here - not fatal anyway.. */
	setsockopt(listen_sock,
		   SOL_SOCKET, SO_REUSEADDR, (void *) &one, sizeof(one));

	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;

	if (server_addr == NULL || strcmp(server_addr, "*") == 0)
	    sin.sin_addr.s_addr = INADDR_ANY;
	else
	    get_addr(server_addr, &sin.sin_addr);
	
	sin.sin_port = htons(server_port);
	
	if (bind(listen_sock, (struct sockaddr *) &sin, sizeof(sin)))
	{
	    syslog(LOG_ERR, "bind() failed: %m");
	    
	    fprintf(stderr, "%s: bind(..., port=%d): ", argv[0], server_port);
	    perror("");
	    exit(1);
	}

        if (listen(listen_sock, n_listen))
        {
            syslog(LOG_ERR, "listen() failed: %m");
	
	    fprintf(stderr, "%s: listen(..., %d): ", argv[0], n_listen);
   	    perror("");
	    exit(1);
        }
    }

    if (getuid() == 0)
    {
	if (pwp)
	{
	    if (initgroups(pwp->pw_name, pwp->pw_gid) < 0)
	    {
		syslog(LOG_ERR, "initgroups(\"%s\", %d) failed: %m",
		       pwp->pw_name, pwp->pw_gid);
		exit(1);
	    }
	}

	if (server_gid != -1)
	    if (setegid(server_gid) < 0)
	    {
		syslog(LOG_ERR, "setegid(%d) failed: %m", server_gid);
		
		fprintf(stderr, "%s: setegid(%d): ", argv[0], server_gid);
		perror("");
		exit(1);
	    }
	
	if (server_uid != -1)
	    if (seteuid(server_uid) < 0)
	    {
		syslog(LOG_ERR, "seteuid(%d) failed: %m", server_uid);
		
		fprintf(stderr, "%s: seteuid(%d): ", argv[0], server_uid);
		perror("");
		exit(1);
	    }
    }
    
    if (!debug)
    {
	int pid;

	for (i = 0; i < 2; i++)
	    if (i != listen_sock)
		s_close(i);
	
	pid = fork();
	if (pid < 0)
	{
	    syslog(LOG_ERR, "fork() failed: %m");
	    exit(1);
	}
	else if (pid > 0)
	    exit(0);

	setsid();
    }
    
    if (debug < 0)
	debug = 0;
    
    syslog(LOG_CRIT, "restarted");
    
    md_init();
    clist_init();
    
    fscache_init();
    urlcache_init();
    hostcache_init();
    usercache_init();
    
    ciptab = s_malloc(max_fds * sizeof(struct connectioninfo *));
    
    log_init(logfile_path, cntfile_path);
    stderr_init(errfile_path);

    if (pidfile_path)
    {
	int fd;


	fd = fd_open(pidfile_path, O_WRONLY+O_CREAT+O_TRUNC, 0644);
	if (fd < 0)
	    syslog(LOG_ERR,
		   "fd_open(\"%s\", O_WRONLY+O_CREAT+O_TRUNC) failed: %m",
		   pidfile_path);
	else
	{
	    fd_printf(fd, "%d\n", getpid());
	    fd_close(fd);
	}
    }

    if (n_processes > 0)
    {
	if (debug > 1)
	    fprintf(stderr, "accept_lock_p(PROCESS) set up\n");
    
	accept_lock_p = s_mutex_create(USYNC_PROCESS);
    }
    else if (n_accept > 1)
    {
	if (debug > 1)
	    fprintf(stderr, "accept_lock_p(THREAD) set up\n");
    
	accept_lock_p = s_mutex_create(USYNC_THREAD);
    }
	
    if (n_processes > 0)
    {
	int child_status;


	if (debug > 1)
	    fprintf(stderr, "creating subprocesses(%d)\n", n_processes);
	
	i = 0;
	
      Loop:
	for (; i < n_processes; i++)
	{
	    int pid;

	    
	    pid = fork();
	    if (pid < 0)
	    {
		syslog(LOG_ERR, "fork() failed: %m");
		if (debug)
		{
		    fprintf(stderr, "%s: fork(): ", argv[0]);
		    perror("");
		}
		exit(1);
	    }
	    if (pid == 0)
		goto DoRequests;
	}

	s_waitpid(-1, &child_status, 0);
	syslog(LOG_NOTICE, "subserver died");
	/* Should restart a new process if some of the old ones die */
	goto Loop;
    }
    
    
  DoRequests:

    actb.sa_handler = SIG_IGN;
    actb.sa_flags = 0;
    sigemptyset(&actb.sa_mask);
    sigaction(SIGPIPE, &actb, NULL);

    actb.sa_handler = SIG_IGN;
    actb.sa_flags = 0;
    sigemptyset(&actb.sa_mask);
    sigaction(SIGCHLD, &actb, NULL);

    actb.sa_handler = sighup_handler;
    actb.sa_flags = 0;
    sigemptyset(&actb.sa_mask);
    sigaction(SIGHUP, &actb, NULL);

    actb.sa_handler = sigusr1_handler;
    actb.sa_flags = 0;
    sigemptyset(&actb.sa_mask);
    sigaction(SIGUSR1, &actb, NULL);

    actb.sa_handler = sigusr2_handler;
    actb.sa_flags = 0;
    sigemptyset(&actb.sa_mask);
    sigaction(SIGUSR2, &actb, NULL);

    actb.sa_handler = sigterm_handler;
    actb.sa_flags = 0;
    sigemptyset(&actb.sa_mask);
    sigaction(SIGTERM, &actb, NULL);

    if (thr_setconcurrency(concurrency))
	syslog(LOG_WARNING, "thr_setconcurrency() failed: %m");

    if (thr_create(NULL,
		   stack_size,
		   (void *(*)(void *)) reconfig_thread,
		   NULL,
		   THR_DETACHED+THR_DAEMON,
		   NULL))
    {
	syslog(LOG_ERR, "thr_create(reconfig_thread) failed: %m");
	exit(1);
    }
    
    if (gc_sleep && gc_time)
    {
	if (thr_create(NULL,
		       stack_size,
		       (void *(*)(void *)) gc_thread,
		       NULL,
		       THR_DETACHED+THR_DAEMON+THR_BOUND,
		       NULL))
	{
	    syslog(LOG_ERR, "thr_create(gc_thread) failed: %m");
	    exit(1);
	}
    }
	
    if (timeout > 0)
    {
	if (thr_create(NULL,
		       stack_size,
		       (void *(*)(void *)) timeout_thread,
		       (void *) &timeout,
		       THR_DETACHED+THR_DAEMON+THR_BOUND,
		       NULL))
	{
	    syslog(LOG_ERR, "thr_create(timeout_thread) failed: %m");
	    exit(1);
	}
    }

    if (debug > 1)
	fprintf(stderr, "creating LESS handler(s)\n");
    
    start_less();

    if (debug > 1)
	fprintf(stderr, "creating accept threads\n");
    
    for (i = 1; i < n_accept; i++)
    {
	if (thr_create(NULL,
		       stack_size,
		       (void *(*)(void *)) accept_thread,
		       (void *) &listen_sock,
		       THR_DETACHED+THR_DAEMON+THR_BOUND,
		       NULL))
	{
	    syslog(LOG_ERR, "thr_create(accept_thread) failed: %m");
	    exit(1);
	}
    }
    
    accept_thread(&listen_sock);

    exit(1);
}
